@@ -1,4 +1,8 @@
-{-# LANGUAGE OverloadedStrings, DeriveDataTypeable, RecordWildCards, DeriveGeneric, CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveGeneric #-}
+{-# LANGUAGE LambdaCase #-}
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE RecordWildCards #-}
 
 module Web.Twitter.Types
        ( UserId
@@ -21,6 +25,8 @@ module Web.Twitter.Types
        , List(..)
        , Entities(..)
        , EntityIndices
+       , ExtendedEntities(..)
+       , ExtendedEntity(..)
        , Entity(..)
        , HashTagEntity(..)
        , UserEntity(..)
@@ -33,24 +39,25 @@ module Web.Twitter.Types
        , Contributor(..)
        , UploadedMedia (..)
        , ImageSizeType (..)
+       , DisplayTextRange (..)
        , checkError
        , twitterTimeFormat
        )
        where
 
 import Control.Applicative
+import Control.Monad
 import Data.Aeson
 import Data.Aeson.Types (Parser)
 import Data.Data
 import Data.HashMap.Strict (HashMap, fromList, union)
+import Data.Int
+import Data.Ratio
 import Data.Text (Text, unpack, pack)
+import Data.Text.Read (decimal)
+import Data.Time
+import Data.Time.Clock.POSIX
 import GHC.Generics
-#if MIN_VERSION_time(1, 5, 0)
-import Data.Time
-#else
-import Data.Time
-import System.Locale
-#endif
 
 newtype TwitterTime = TwitterTime { fromTwitterTime :: UTCTime }
 
@@ -83,16 +90,12 @@ twitterTimeFormat = "%a %b %d %T %z %Y"
 
 instance FromJSON TwitterTime where
     parseJSON = withText "TwitterTime" $ \t ->
-#if MIN_VERSION_time(1, 5, 0)
         case parseTimeM True defaultTimeLocale twitterTimeFormat (unpack t) of
-#else
-        case parseTime defaultTimeLocale twitterTimeFormat (unpack t) of
-#endif
             Just  d -> pure $ TwitterTime d
             Nothing -> fail $ "Could not parse twitter time. Text was: " ++ unpack t
 
 instance ToJSON TwitterTime where
-    toJSON t = String $ pack $ formatTime defaultTimeLocale twitterTimeFormat $ fromTwitterTime t 
+    toJSON t = String $ pack $ formatTime defaultTimeLocale twitterTimeFormat $ fromTwitterTime t
 
 instance FromJSON StreamingAPI where
     parseJSON v@(Object o) =
@@ -125,7 +128,7 @@ data Status = Status
     , statusCreatedAt :: UTCTime
     , statusCurrentUserRetweet :: Maybe StatusId
     , statusEntities :: Maybe Entities
-    , statusExtendedEntities :: Maybe Entities
+    , statusExtendedEntities :: Maybe ExtendedEntities
     , statusFavoriteCount :: Integer
     , statusFavorited :: Maybe Bool
     , statusFilterLevel :: Maybe Text
@@ -149,6 +152,7 @@ data Status = Status
     , statusWithheldCopyright :: Maybe Bool
     , statusWithheldInCountries :: Maybe [Text]
     , statusWithheldScope :: Maybe Text
+    , statusDisplayTextRange :: Maybe DisplayTextRange
     } deriving (Show, Eq, Data, Typeable, Generic)
 
 instance FromJSON Status where
@@ -176,12 +180,13 @@ instance FromJSON Status where
                <*> o .:? "retweeted"
                <*> o .:? "retweeted_status"
                <*> o .:  "source"
-               <*> o .:  "text"
+               <*> (o .: "full_text" <|> o .: "text")
                <*> o .:  "truncated"
                <*> o .:  "user"
                <*> o .:? "withheld_copyright"
                <*> o .:? "withheld_in_countries"
                <*> o .:? "withheld_scope"
+               <*> o .:? "display_text_range"
     parseJSON v = fail $ "couldn't parse status from: " ++ show v
 
 instance ToJSON Status where
@@ -216,6 +221,7 @@ instance ToJSON Status where
                                , "withheld_copyright"       .= statusWithheldCopyright
                                , "withheld_in_countries"    .= statusWithheldInCountries
                                , "withheld_scope"           .= statusWithheldScope
+                               , "display_text_range"       .= statusDisplayTextRange
                                ]
 
 data SearchResult body =
@@ -342,46 +348,54 @@ instance ToJSON RetweetedStatus where
                                         , "coordinates"         .= rsCoordinates
                                         ]
 
-data DirectMessage =
-    DirectMessage
-    { dmCreatedAt          :: UTCTime
-    , dmSenderScreenName   :: Text
-    , dmSender             :: User
-    , dmText               :: Text
-    , dmRecipientScreeName :: Text
-    , dmId                 :: StatusId
-    , dmRecipient          :: User
-    , dmRecipientId        :: UserId
-    , dmSenderId           :: UserId
-    , dmCoordinates        :: Maybe Coordinates
+type EventId = Integer
+
+data DirectMessage = DirectMessage
+    { dmId :: EventId
+    , dmCreatedTimestamp :: UTCTime
+    , dmTargetRecipientId :: UserId
+    , dmSenderId :: UserId
+    , dmText :: Text
+    , dmEntities :: Entities
     } deriving (Show, Eq, Data, Typeable, Generic)
 
+parseIntegral :: Integral a => Text -> Parser a
+parseIntegral v = either (\_ -> fail $ "couldn't parse stringized int: " ++ show v) (return . fst) $ decimal v
+
+epochMsToUTCTime :: Int64 -> UTCTime
+epochMsToUTCTime = posixSecondsToUTCTime . fromRational . (% 1000) . fromIntegral
+
+parseUnixTimeString :: Text -> Parser UTCTime
+parseUnixTimeString = fmap epochMsToUTCTime <$> parseIntegral
+
+unixTimeToEpochInt :: UTCTime -> Int
+unixTimeToEpochInt = floor . (* 1000) . utcTimeToPOSIXSeconds
+
 instance FromJSON DirectMessage where
-    parseJSON (Object o) = checkError o >>
-        DirectMessage <$> (o .:  "created_at" >>= return . fromTwitterTime)
-                      <*> o .:  "sender_screen_name"
-                      <*> o .:  "sender"
-                      <*> o .:  "text"
-                      <*> o .:  "recipient_screen_name"
-                      <*> o .:  "id"
-                      <*> o .:  "recipient"
-                      <*> o .:  "recipient_id"
-                      <*> o .:  "sender_id"
-                      <*> o .:? "coordinates"
-    parseJSON v = fail $ "couldn't parse direct message from: " ++ show v
+    parseJSON (Object o) = do
+        _ <- checkError o
+        messageCreate <- o .: "message_create"
+        messageData <- messageCreate .: "message_data"
+
+        DirectMessage
+            <$> (o .: "id" >>= parseIntegral)
+            <*> (o .: "created_timestamp" >>= parseUnixTimeString)
+            <*> (messageCreate .: "target" >>= (.: "recipient_id") >>= parseIntegral)
+            <*> (messageCreate .: "sender_id" >>= parseIntegral)
+            <*> messageData .: "text"
+            <*> messageData .: "entities"
+    parseJSON v = fail $ "couldn't parse direct message create event from: " ++ show v
 
 instance ToJSON DirectMessage where
-    toJSON DirectMessage{..} = object [ "created_at"            .= TwitterTime dmCreatedAt
-                                      , "sender_screen_name"    .= dmSenderScreenName
-                                      , "sender"                .= dmSender
-                                      , "text"                  .= dmText
-                                      , "recipient_screen_name" .= dmRecipientScreeName
-                                      , "id"                    .= dmId
-                                      , "recipient"             .= dmRecipient
-                                      , "recipient_id"          .= dmRecipientId
-                                      , "sender_id"             .= dmSenderId
-                                      , "coordinates"           .= dmCoordinates
-                                      ]
+    toJSON DirectMessage {..} =
+        object
+            [ "id" .= show dmId
+            , "created_timestamp" .= show (unixTimeToEpochInt dmCreatedTimestamp)
+            , "message_create" .= object ["message_data" .= object ["text" .= dmText, "entities" .= dmEntities]
+                                         , "target" .= object ["recipient_id" .= show dmTargetRecipientId]
+                                         , "sender_id" .= show dmSenderId
+                                         ]
+            ]
 
 data EventType = Favorite | Unfavorite
                | ListCreated | ListUpdated | ListMemberAdded
@@ -459,6 +473,7 @@ data User = User
     , userDefaultProfile :: Bool
     , userDefaultProfileImage :: Bool
     , userDescription :: Maybe Text
+    , userEmail :: Maybe Text
     , userFavoritesCount :: Int
     , userFollowRequestSent :: Maybe Bool
     , userFollowing :: Maybe Bool
@@ -467,7 +482,7 @@ data User = User
     , userGeoEnabled :: Bool
     , userId :: UserId
     , userIsTranslator :: Bool
-    , userLang :: LanguageCode
+    , userLang :: Maybe LanguageCode
     , userListedCount :: Int
     , userLocation :: Maybe Text
     , userName :: Text
@@ -503,6 +518,7 @@ instance FromJSON User where
              <*> o .:  "default_profile"
              <*> o .:  "default_profile_image"
              <*> o .:? "description"
+             <*> fmap join (o .:? "email") -- The field can be a null value
              <*> o .:  "favourites_count"
              <*> o .:? "follow_request_sent" .!= Nothing
              <*> o .:? "following" .!= Nothing
@@ -546,6 +562,7 @@ instance ToJSON User where
                              , "default_profile"                    .= userDefaultProfile
                              , "default_profile_image"              .= userDefaultProfileImage
                              , "description"                        .= userDescription
+                             , "email"                              .= userEmail
                              , "favourites_count"                   .= userFavoritesCount
                              , "follow_request_sent"                .= userFollowRequestSent
                              , "following"                          .= userFollowing
@@ -828,6 +845,7 @@ instance ToJSON Entities where
                                  , "media"          .= enMedia
                                  ]
 
+
 -- | The character positions the Entity was extracted from
 --
 --   This is experimental implementation.
@@ -851,6 +869,65 @@ instance ToJSON a => ToJSON (Entity a) where
                             (Object o) -> Object $ union o $ fromList [("indices"::Text, toJSON entityIndices)]
                             _          -> error "Entity body must produce an object."
 
+data ExtendedEntities =
+  ExtendedEntities {
+    exeMedia :: [Entity ExtendedEntity]
+  } deriving (Show, Eq, Data, Typeable, Generic)
+
+instance FromJSON ExtendedEntities where
+  parseJSON (Object o) =
+    ExtendedEntities <$> o .:? "media" .!= []
+
+  parseJSON v = fail $ "couldn't parse extended entity from: " ++ show v
+
+instance ToJSON ExtendedEntities where
+    toJSON ExtendedEntities{..} = object [ "media" .= exeMedia ]
+
+
+-- Extended entities are like entities, but contain special media features like
+-- video or multiple photos
+data ExtendedEntity =
+  ExtendedEntity
+  {
+    exeID :: StatusId,
+    exeMediaUrl :: URIString,
+    exeMediaUrlHttps :: URIString,
+    exeSizes :: HashMap Text MediaSize,
+    exeType :: Text,
+    -- exeVideoInfo :: ???
+    exeDurationMillis :: Maybe Double,
+    -- exeVariants :: ??
+    exeExtAltText :: Maybe String,
+    exeURL :: URLEntity
+  } deriving (Show, Eq, Data, Typeable, Generic)
+
+instance FromJSON ExtendedEntity where
+  parseJSON v@(Object o) =
+    ExtendedEntity <$> o .:  "id"
+                   <*> o .:  "media_url"
+                   <*> o .:  "media_url_https"
+                   <*> o .:  "sizes"
+                   <*> o .:  "type"
+                   <*> o .:? "duration_millis"
+                   <*> o .:?  "ext_alt_text"
+                   <*> parseJSON v
+
+  parseJSON v = fail $ "couldn't parse extended entity from:" ++ show v
+
+instance ToJSON ExtendedEntity where
+  toJSON ExtendedEntity{..} = object [ "id"                .= exeID
+                                     , "media_url"       .= exeMediaUrl
+                                     , "media_url_https" .= exeMediaUrlHttps
+                                     , "sizes"           .= exeSizes
+                                     , "type"            .= exeType
+                                     , "duration_millis" .= exeDurationMillis
+                                     , "ext_alt_text"    .= exeExtAltText
+                                     , "url"             .= ueURL exeURL
+                                     , "expanded_url"    .= ueExpanded exeURL
+                                     , "display_url"     .= ueDisplay exeURL
+                                     ]
+
+
 data Contributor = Contributor
     { contributorId :: UserId
     , contributorScreenName :: Maybe Text
@@ -909,3 +986,18 @@ instance ToJSON UploadedMedia where
                                       , "size"      .= uploadedMediaSize
                                       , "image"     .= uploadedMediaImage
                                       ]
+
+-- | unicode code point indices, identifying the inclusive start and exclusive end of the displayable content of the Tweet.
+data DisplayTextRange = DisplayTextRange
+    { displayTextRangeStart :: Int
+    , displayTextRangeEnd :: Int -- ^ exclusive
+    } deriving (Show, Eq, Ord, Data, Typeable, Generic)
+
+instance FromJSON DisplayTextRange where
+    parseJSON v = do
+        parseJSON v >>= \case
+            [s, e] -> pure $ DisplayTextRange s e
+            unexpected ->
+                fail $ "parsing DisplayTextRange failed, expected [Int, Int], but got: " ++ show unexpected
+instance ToJSON DisplayTextRange where
+    toJSON (DisplayTextRange s e) = toJSON [s, e]
